import numpy as np

def ensure_np_array(value, dtype=np.float32):
    '''
    function that returns the varable as is, or casts to desired dtype'''
    if not isinstance(value, np.ndarray):
        as_np_array = np.array([value], dtype=dtype)
        return as_np_array
    if (value.dtype != dtype):
        print(f"in online_data_structures: casting to type {dtype} from {value.dtype}")
        return value.astype(dtype)
    
    return value

class NpVec:
    '''
    makes all data to 1d array and appends to existing vector
    if max_size is not None, then this is a circular buffer'''
    def __init__(self, dtype=np.float32, max_size = None):
       self.dtype = dtype
       self.max_size = max_size
       self.init()
    
    def append(self, value):
        # cast to dtype if necessary and flatten
        value_np = ensure_np_array(value, dtype=self.dtype).reshape(-1)
        value_np_size = value_np.size
        
        
        if self.max_size is None:
            while value_np_size + self.size > self.capacity:
                self.increase_capacity()
        
            start = self.size
            end = self.size + value_np_size
            self.data[start:end] = value_np
        
            
        
        else:
            if value_np_size > self.max_size:
                print(f"Error in data structure, data that is added/appended has size {value_np_size}, however, the structure was initialized with a max size i.e. capacity of {self.max_size}. can not add this data. aborting.")
                return
        
            # wrap around?
            if value_np_size + self.current_write_pos > self.capacity:
                # first part:
                first_part_len = self.capacity-self.current_write_pos
                start = self.current_write_pos
                end = self.capacity
                self.data[start:end] = value_np[:first_part_len]

                # second part
                second_part_len = value_np_size-first_part_len
                start = 0
                end = second_part_len
                self.data[start:end] = value_np[first_part_len:]
                self.current_write_pos = second_part_len

                

            else:
                start = self.current_write_pos
                end = self.current_write_pos + value_np_size
                
                
                self.data[start:end] = value_np
                self.current_write_pos += value_np_size
                

        self.size += value_np_size


    def add(self, value):
        self.append(value)

    def get(self):
        if self.max_size is None:
            return self.data[:self.size]
        
        # wrap around not yet achieved:
        if self.size < self.capacity:
            return self.data[:self.size]
        
        # wrap around:
        res = np.zeros_like(self.data)

        # first part
        start = self.current_write_pos
        first_part_len = self.capacity - self.current_write_pos
        res[:first_part_len] = self.data[start:]
        res[first_part_len:] = self.data[:start]
        return res
    
    def get_all(self):
        return self.get()

    def get_size(self):
        return self.size
    
    def get_capacity(self):
        return self.capacity
    
    def clear(self):
        self.init()

    # private
    def init(self):
        self.size = 0
        if self.max_size is None:
            self.capacity = 10
        else:
            self.capacity = self.max_size
            self.current_write_pos = 0
            
        self.data = np.zeros((self.capacity,), dtype=self.dtype)

    def increase_capacity(self):
        # make new memort of double size
        new_data = np.zeros((2*self.capacity,), dtype=self.dtype)
        # copy old data over
        new_data[:self.size] = self.data[:self.size] 
        # increase capacity
        self.capacity = 2*self.capacity
        self.data = new_data
    
    def get_schema_injection(self):
        pass

    
        
